from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Literal, NewType

from wattro_sync.api.sqlite_api import SQLiteSyncInfo

FieldMapping = NewType("FieldMapping", dict[str, dict[str, None | int | str]])


@dataclass
class ConnectionStructure:
    connection_type: Literal["SQLite"]
    sync_info: SQLiteSyncInfo
    field_mapping: FieldMapping
    encoding: str = "utf-8"


@dataclass
class WattroCfg:
    domain: str
    api_key: str


@dataclass
class MailCfg:
    api_key: str
    to_emails: str
    log_level: int = logging.WARNING
    form_email: str = "admin@wattro.de"


@dataclass
class SyncCfg:
    wattro_cfg: WattroCfg
    asset: None | ConnectionStructure
    project: None | ConnectionStructure
    mail_cfg: None | MailCfg


class ConfigDegenerated(Exception):
    """Config File exists but is not as expected"""



from __future__ import annotations

import abc
import json
import logging
import urllib.error
import urllib.parse
import urllib.request

import requests


class RESTApi(abc.ABC):
    def _get_headers(self) -> dict:
        headers = getattr(self, "headers", None)
        if headers is None:
            raise NotImplementedError("headeres not set")
        return headers

    def _get_hostname(self) -> str:
        hostname = getattr(self, "hostname", None)
        if hostname is None:
            raise NotImplementedError("hostname not set")
        return hostname

    def _get_protocol(self) -> str:
        return getattr(self, "protocol", "https")

    def _get_base_waittime(self) -> float:
        return getattr(self, "base_waittime", 3.0)

    def _request(
        self,
        method: str,
        url: str,
        *,
        custom_header: dict | None = None,
        data: dict | None = None,
        params: dict | None = None,
    ) -> dict:
        header = self._get_headers()
        if custom_header:
            header.update(custom_header)
        parsed_data = None
        if data is not None:
            parsed_data = json.dumps(data)
            header.update({"Content-Type": "application/json"})
        res = requests.request(
            method=method,
            url=url,
            data=parsed_data,
            params=params,
            headers=header,
            timeout=20,
        )
        if res.ok:
            return res.json()
        logging.error("%s Request to %s Failed", method, url)
        logging.error("%s: %s", res.status_code, res.content)
        raise ConnectionError()

    def _url(self, path: str) -> str:
        return urllib.parse.urlunparse(
            (self._get_protocol(), self._get_hostname(), path, None, None, None)
        )

    def _get(self, path: str, params: dict | None = None) -> dict:
        return self._request(method="GET", url=self._url(path), params=params)

    def _post(self, path: str, data: dict) -> dict:
        return self._request(method="POST", data=data, url=self._url(path))


class WattroNodeApi(RESTApi):
    def __init__(self, domain: str, api_key: str):
        if domain == "local":
            self.protocol = "http"
            self.hostname = "127.0.0.1:8000"
        else:
            self.protocol = "https"
            self.hostname = f"node.{domain}.wattro.de"
        self.headers = {"Authorization": f"Api-Key {api_key}"}

    def _get(self, path: str, params: dict | None = None) -> dict:
        path = path.strip("/") + "/"
        return super()._get(path, params)

    def get_ref_url(self) -> str:
        return self._url("healthchecks/")

    def get_health(self) -> dict:
        return self._get("healthchecks/")

    def get_assets(self, params: dict | None = None) -> dict:
        return self._get("node/asset/", params)

    def get_idents(self, target: str) -> list[str]:
        res = self._get(f"/sync/{target}/get_idents/")
        if "idents" not in res:
            raise RuntimeError(f"Unexpected API Result {res}")
        idents = res["idents"]
        if not isinstance(idents, list):
            RuntimeError(f"Expected list from Wattro API got {idents!r}")
        return idents

    def get_fields(self, path: str) -> dict:
        res = self._request(method="OPTIONS", url=self._url(path))
        return res["actions"]["POST"]

    @classmethod
    def get_healthy_api(cls, domain: str, api_key: str) -> WattroNodeApi:
        api = cls(domain, api_key)
        try:
            health = api.get_health()
        except urllib.error.URLError as urllib_err:
            raise ConnectionError(
                f"Failed to connect to {api.get_ref_url()!r}"
            ) from urllib_err

        if not health["auth_status"]["has_permission"]:
            raise ConnectionError(
                f"Not allowed to connect to {api.get_ref_url()!r}: {health}"
            )
        return api

    def bulk_create(self, target: str, new_target_data: list[dict]) -> None:
        self._post(f"/sync/{target}/bulk/", data={"new_data": new_target_data})

    def update_by_ident(self, target: str, new_target_data: dict) -> None:
        self._post(
            f"/sync/{target}/update_by_ident/", data={"new_data": new_target_data}
        )



import logging
from typing import overload, Literal


class StupidTerminalChoice:
    def __init__(self, menu_entries: list, **kwargs):
        self.menu_entries = menu_entries
        self.kwargs = kwargs

    def _is_multi(self) -> bool:
        key = "multi_select"
        return self.kwargs.get(key, False)

    def show(self) -> int | tuple[int, ...] | None:
        print(f"\n\n##\n {self.kwargs.get('title', 'Auswahl:')}")
        print("[q]\t keins davon")
        for i, entry in enumerate(self.menu_entries):
            print(f"[{i}]\t {entry}")
        if self._is_multi():
            choices = input("Auswahl (etwa 0 2 3):\n").split(" ")
            if len(choices) == 1 and choices[0] == "q":
                return None
            return tuple(int(x) for x in choices if x != "")
        # else
        choice = input("Auswahl (etwa 0)\n").strip()
        if choice == "q":
            return None
        return int(choice)


try:
    from simple_term_menu import TerminalMenu
except NotImplementedError as platform_err:
    TerminalMenu = StupidTerminalChoice
    logging.warning(f"{platform_err} - falling back to {TerminalMenu!r}")

from wattro_sync.api.api_mapping import ApiNameToStructureMapping

SOURCE_CHOICES = ApiNameToStructureMapping.keys()
TARGET_NODE_MAPPING = {"project": "/project/project/", "asset": "/node/asset/"}


# TerminalMenu Helpers
@overload
def multi_select(
    menu_entries: list[str], required: Literal[True], **kwargs
) -> tuple[int]:
    ...


@overload
def multi_select(
    menu_entries: list[str], required: Literal[False], **kwargs
) -> None | tuple[int]:
    ...


def multi_select(
    menu_entries: list[str], required: bool = False, **kwargs
) -> None | tuple[int]:
    is_multi = kwargs.pop("multi_select", True)
    res = _select(menu_entries, is_multi, required, **kwargs)
    if isinstance(res, int):
        raise RuntimeError("Failed to parse mutli select", res)
    return res


@overload
def select(menu_entries: list[str], required: Literal[True], **kwargs) -> int:
    ...


@overload
def select(menu_entries: list[str], required: Literal[False], **kwargs) -> None | int:
    ...


def select(menu_entries: list[str], required: bool = False, **kwargs) -> None | int:
    is_multi = kwargs.pop("multi_select", False)
    res = _select(menu_entries, is_multi, required, **kwargs)
    if isinstance(res, tuple):
        raise RuntimeError("Failed to parse single value", res)
    return res


def _select(
    menu_entries: list[str], is_multi: bool, required: bool, **kwargs
) -> None | int | tuple[int]:
    is_many = len(menu_entries) > 7
    res = TerminalMenu(
        menu_entries,
        **kwargs,
        multi_select=is_multi,
        show_multi_select_hint=is_multi,
        show_multi_select_hint_text="<Leertaste> um an/abzuwählen, <Enter> um anzuwählen und abzuschließen.",
        show_search_hint=is_many and not is_multi,
        show_search_hint_text="'/' Eingeben um Filter zu starten",
    ).show()
    if required and (res is None or res == []):
        input("Pflichtfeld. (Enter drücken um erneut zu wählen).")
        return _select(menu_entries, is_multi, required, **kwargs)
    return res



import logging
import sys

from wattro_sync.file_access.read_write import get_base_folder_path

formatter = logging.Formatter("%(asctime)s | %(levelname)s | %(message)s")


def get_file_handler() -> logging.FileHandler:
    wattro_folder = get_base_folder_path()
    logging_fh = logging.FileHandler(wattro_folder / "logs.log", mode="a")
    logging_fh.setFormatter(formatter)
    logging_fh.setLevel(logging.DEBUG)
    return logging_fh


def get_stdout_handler() -> logging.StreamHandler:
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.DEBUG)
    stdout_handler.setFormatter(formatter)
    return stdout_handler



from __future__ import annotations

import json
import logging
import os
import pathlib
import typing

from wattro_sync.config_reader.types import ConfigDegenerated

BASE_FOLDER = ".wattro_sync"
FILE_NAMES = ["cfg.json", "history.json"]
CON_TYPE_KEY = "connection_type"
CON_INFO_KEY = "connection_info"
FIELD_MAP_KEY = "field_mapping"

ShortType = typing.Literal["cfg", "history"]


def exists(file_type: ShortType) -> bool:
    fp = _get_file_path(file_type)
    return fp.exists()


def get_or_create(file_type: ShortType) -> tuple[pathlib.Path, bool]:
    """
    file_path, created = get_or_create('cfg')
    """
    created = False
    file_path = _get_file_path(file_type)
    if not file_path.is_file():
        created = True
        write_path(file_path, {})
    return file_path, created


def write_path(file_path: pathlib.Path, val: typing.Any, pretty: bool = False) -> None:
    if file_path.exists():
        logging.info("Overwriting exsiting file %s", file_path)
    txt = ""
    if pretty:
        txt = json.dumps(val, indent=4, sort_keys=True)
    else:
        txt = json.dumps(val)
    file_path.write_text(txt)


def write(file_type: ShortType, val: typing.Any) -> None:
    file_p, _ = get_or_create(file_type)
    return write_path(file_p, val)


def read_path(file_path: pathlib.Path) -> dict:
    try:
        raw_read = json.loads(file_path.read_text())
    except (json.decoder.JSONDecodeError, AssertionError) as mal_cfg:
        raise ConfigDegenerated(f"{file_path} could not be parsed.") from mal_cfg
    if not isinstance(raw_read, dict):
        raise ConfigDegenerated(f"{file_path} not a dict.")
    return raw_read


def read(file_type: ShortType) -> dict:
    file_p, _ = get_or_create(file_type)
    return read_path(file_p)


def update(file_type: ShortType, target: str, key: str, val: typing.Any) -> None:
    file_p, _ = get_or_create(file_type)
    file_dict = read_path(file_p)
    if target in file_dict:
        if key in file_dict[target]:
            logging.warning(
                "Overwriting existing value for '%s' in '%s'. Old value: '%s'. New value: %s",
                key,
                target,
                file_dict[target][key],
                val,
            )
    else:
        file_dict[target] = {}
    file_dict[target].update({key: val})
    write_path(file_p, file_dict, pretty=(file_type == "cfg"))


def hist_update(target: str, key: str, val: typing.Any) -> None:
    return update("history", target, key, val)


def cfg_update(target: str, key: str, val: typing.Any) -> None:
    return update("cfg", target, key, val)


def cfg_rm() -> None:
    """removes the config file"""
    file_path = _get_file_path("cfg")
    if not file_path.is_file():
        logging.info("Nothing to do.")
        return
    os.unlink(file_path)


def get_base_folder_path() -> pathlib.Path:
    """Path where all configs and logs are stored"""
    return pathlib.Path.home() / BASE_FOLDER


def _get_or_create_base_folder() -> pathlib.Path:
    """gets or creates wattro config folder"""
    wattro_folder = get_base_folder_path()
    if not wattro_folder.is_dir():
        logging.info("Erzeuge Wattro Konfig Ordner: %s", wattro_folder)
        os.mkdir(wattro_folder)
    return wattro_folder


def _get_file_path(file_type: ShortType) -> pathlib.Path:
    folder = _get_or_create_base_folder()
    file_name = [x for x in FILE_NAMES if x.startswith(file_type)]
    if not len(file_name) == 1:
        raise FileNotFoundError(f"{file_type!r} is no valid selection.")
    return folder / file_name.pop()



from __future__ import annotations

from wattro_sync.api.odbc_api import OdbcSyncInfo, OdbcSrcCliAltSample


class TopKontorSyncInfo(OdbcSyncInfo):
    """branded OdbcSyncInfo"""


class TopKontorApi(OdbcSrcCliAltSample):
    @classmethod
    def get_collections(cls, connection_info: str) -> list[str]:
        raise NotImplementedError("Not implemented for TopKontor.")

    @classmethod
    def get_fields(
        cls, connection_info: str, collection: str
    ) -> tuple[list[str], dict[str, list]]:
        raise NotImplementedError("Not implemented for TopKontor.")



import dataclasses
from typing import Any, Sequence

from wattro_sync.api.odbc_api import OdbcSyncInfo, OdbcSrcCliAltSample
from wattro_sync.api.src_cli import CollectionInfo, DBRes


@dataclasses.dataclass
class MosaikConInfo:
    Server: str
    Database: str
    Trusted_Connection: str = "Yes"
    Driver: str = "{SQL Server}"

    def to_connection_str(self) -> str:
        return ";".join(f"{k}={v}" for k, v in dataclasses.asdict(self).items())


class MosaikSyncInfo(OdbcSyncInfo):
    """branded ODBC Sync Info"""


class MosaikApi(OdbcSrcCliAltSample):
    MAX_SQL_TOKENS = 2_000

    @classmethod
    def get_collections(cls, connection_info: Any) -> list[str]:
        fake_api = cls(
            MosaikSyncInfo(
                odbc_connection_str=connection_info,
                collection_info=CollectionInfo.empty(),
            )
        )
        db_res = fake_api._exec(
            "SELECT name FROM SYSOBJECTS WHERE xtype='U' OR xtype='V';"
        )
        return sorted([x["name"] for x in db_res])

    @classmethod
    def get_fields(
        cls, connection_info: Any, collection: str
    ) -> tuple[list[str], dict[str, list]]:
        fake_api = cls(
            MosaikSyncInfo(
                odbc_connection_str=connection_info,
                collection_info=CollectionInfo.empty(),
            )
        )
        meta_info = fake_api._exec(
            "SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS "
            f"WHERE TABLE_NAME = '{collection}';"
        )
        field_names = sorted(list(str(x["COLUMN_NAME"]) for x in meta_info))

        db_res = fake_api._exec(f"SELECT TOP 50 * FROM {collection}")
        transposed_rows = list(map(list, zip(*db_res.rows)))
        sample_values = {
            field: val for field, val in zip(db_res.description, transposed_rows)
        }
        return field_names, sample_values

    def get_new(self, known_idents: Sequence[str]) -> DBRes:
        """
        Get all entries on the CollectionInfo collection that are not identified by `known_idents`
        This assumes that we are conntected to a view that sorts by changed date and is limited to the last 2k (or less)
        """
        return super().get_new(known_idents[: self.MAX_SQL_TOKENS])

    def get_old(self, known_idents: Sequence[str]) -> DBRes:
        """
        The DB is limited to 2k SQL tokens, so we have to limit known idents.
        This assumes that we are conntected to a view that sorts by changed date and is limited to the last 2k (or less)
        """
        return super().get_old(known_idents[: self.MAX_SQL_TOKENS])



from dataclasses import dataclass

from .mosaik_api import MosaikApi, MosaikSyncInfo
from .sqlite_api import SQLiteSyncInfo, SQLiteApi
from .src_cli import SrcCli, SyncInfo
from .topkontor_api import TopKontorSyncInfo, TopKontorApi


@dataclass
class ApiStructure:
    connection_info: type[SyncInfo]
    api: type[SrcCli]


ApiNameToStructureMapping: dict[str, ApiStructure] = {
    "Mosaik": ApiStructure(MosaikSyncInfo, MosaikApi),
    "TopKontor": ApiStructure(TopKontorSyncInfo, TopKontorApi),
    "Benning": ApiStructure(SQLiteSyncInfo, SQLiteApi),
    "SQLite": ApiStructure(SQLiteSyncInfo, SQLiteApi),
}



import logging

import sendgrid
from sendgrid.helpers import mail

from wattro_sync.config_reader.types import MailCfg


class MailApi:
    def __init__(self, mail_cfg: MailCfg | None):
        if mail_cfg is None:
            self.dummy = True
            self.cfg = MailCfg("", "")
            return
        self.dummy = False
        self.cfg = mail_cfg
        self.sg = sendgrid.SendGridAPIClient(mail_cfg.api_key)

    def send_registered(self) -> bool:
        """returns True if mail was sent successfully."""
        msg = "E-Mail Logging für das Wattro Synchronisations Script wurde aktiviert."
        res = self._send(self._msg("Aktivierung erfolgreich.", msg))
        return res is not None

    def _msg(self, subject: str, html_content: str) -> mail.Mail:
        return mail.Mail(
            from_email=self.cfg.form_email,
            to_emails=self.cfg.to_emails,
            subject=subject,
            html_content=html_content,
        )

    def _send(self, message: mail.Mail) -> None | str:
        """try to send a mail, return None on fail"""
        if self.dummy:
            return None
        try:
            res = self.sg.send(message)
            return res.body.decode("utf-8")
        except Exception as err:
            logging.error(
                f"Sending of {message} failed: {err} | {getattr(err, 'message', '')}"
            )
            return None

    def send(self, log_lvl: int, msg: str) -> None:
        if self.dummy or log_lvl < self.cfg.log_level:
            logging.log(log_lvl, f"(sending mail skipped), {msg}")
            return
        self._send(
            self._msg(
                subject=f"Wattro Sync: {logging.getLevelName(log_lvl)}",
                html_content=msg,
            )
        )



from __future__ import annotations

import dataclasses
import sqlite3

from .src_cli import SrcCli, DBRes, CollectionInfo, SyncInfo


class SQLiteSyncInfo(SyncInfo):
    db_path: str
    collection_info: CollectionInfo

    def __init__(self, db_path: str, collection_info: CollectionInfo):
        self.db_path = db_path
        self.collection_info = collection_info

    def asdict(self) -> dict:
        return {
            "db_path": str(self.db_path),
            "collection_info": dataclasses.asdict(self.collection_info),
        }

    @classmethod
    def from_dict(cls, info: dict) -> SQLiteSyncInfo:
        return cls(
            db_path=info["db_path"],
            collection_info=CollectionInfo(**info["collection_info"]),
        )


class SQLiteApi(SrcCli):
    def __init__(self, sync_info: SQLiteSyncInfo):
        self.db_path = sync_info.db_path
        self.collection_info = sync_info.collection_info

    @classmethod
    def get_fields(
        cls, connection_info: str, collection: str
    ) -> tuple[list[str], dict[str, list]]:
        fake_api = cls(SQLiteSyncInfo(connection_info, CollectionInfo.empty()))
        meta_info = fake_api._exec(f"PRAGMA table_info({collection})")
        field_names = sorted(list([str(x["name"]) for x in meta_info]))

        db_res = fake_api._exec(f"SELECT * FROM {collection} LIMIT 50")
        transposed_rows = list(map(list, zip(*db_res.rows)))
        sample_values = {
            field: val for field, val in zip(db_res.description, transposed_rows)
        }
        return field_names, sample_values

    @classmethod
    def get_collections(cls, connection_info: str) -> list[str]:
        """
        get_collections('/path/to/db')
        """
        fake_api = cls(
            SQLiteSyncInfo(
                db_path=connection_info, collection_info=CollectionInfo.empty()
            )
        )
        db_res = fake_api._exec(f"PRAGMA table_list")
        return sorted([x["name"] for x in db_res])

    def _exec(self, qry: str, params=None) -> DBRes:
        try:
            cnxn = sqlite3.connect(self.db_path)
        except Exception as err:
            raise ConnectionError(f"Failed to open db at {self.db_path}") from err
        try:
            cursr = cnxn.cursor()
            if params is None:
                params = tuple()
            rows = cursr.execute(qry, params).fetchall()
            if not rows:
                res = DBRes([], [])
            else:
                res = DBRes([n[0] for n in cursr.description], rows)
        except Exception as err:
            raise RuntimeError(
                "Failed to execute {qry = !r} with {params = !r}"
            ) from err
        finally:
            cnxn.close()
        return res



#!/bin/env python3
import argparse
import logging
import pathlib
import typing

import wattro_sync.config_reader.access
import wattro_sync.file_access.read_write
from wattro_sync.api.mail import MailApi
from wattro_sync.api.mosaik_api import MosaikSyncInfo, MosaikConInfo
from wattro_sync.config_reader.types import SyncCfg, MailCfg
from wattro_sync.helpers import (
    SOURCE_CHOICES,
    TARGET_NODE_MAPPING,
    multi_select,
    select,
)
from .api.api_mapping import ApiNameToStructureMapping, ApiStructure
from .api.rest_api import WattroNodeApi
from .api.sqlite_api import SQLiteSyncInfo
from .api.src_cli import CollectionInfo, SrcCli, SyncInfo
from .config_reader import access as cfg_access


def main() -> int:
    logging.basicConfig(level=logging.INFO)
    args = parse_args()
    try:
        cfg, wattro_api = retrive_cfg_and_wattro_api()
    except ConnectionError as c_err:
        logging.critical("Verbinundgsfehler: %s", c_err)
        if select(["nein", "ja"], required=True, title="Konfiguration löschen?"):
            logging.info("Lösche Konfigurationsdatei.")
            wattro_sync.file_access.read_write.cfg_rm()
        return -1

    logging.info("Erzeuge Konfiguration für %s -> %s", args.source, args.target)
    try:
        api_struct: ApiStructure = ApiNameToStructureMapping[args.source]
        collection_info = write_source_connection_and_get_info(api_struct, args)
    except KeyError:
        return -1
    except NotImplementedError:
        logging.critical(
            "Automatisches Erzeugen dieser Schnittstelle nicht implementiert."
        )
        return -1

    write_mapping(args, collection_info, wattro_api)
    print(
        f"Synchronisations Konfiguration für {args.target!r} mit Quelle {args.source!r} gespeichert."
    )

    return 0


def write_mapping(
    args: argparse.Namespace, collection_info: CollectionInfo, wattro_api: WattroNodeApi
) -> None:
    raw_fields = wattro_api.get_fields(TARGET_NODE_MAPPING[args.target])
    field_mapping: dict[str, dict[str, None | int | str]] = {}
    src_fields = collection_info.fields
    for key, value in raw_fields.items():
        if value.get("read_only", True):
            logging.info("Überspringe 'read_path only' Feld %s; %s", key, value)
            continue
        if key == "human_id":
            ident = collection_info.ident
            logging.info(f"Setze 'human_id' auf {ident!r}")
            field_mapping[key] = {**value, "src": _make_cfg_field(ident)}
            continue
        field_mapping[key] = {**value, "src": select_src(key, src_fields, value)}
    cfg_access.write_field_mapping(args.target, field_mapping)


def write_source_connection_and_get_info(
    api_struct: ApiStructure, args: argparse.Namespace
) -> CollectionInfo:
    sync_info: SyncInfo
    if api_struct.connection_info == SQLiteSyncInfo:
        con_info = get_sqlite_connection_info()
        collection_info = create_collection_info(api_struct.api, con_info)
        sync_info = SQLiteSyncInfo(con_info, collection_info)
    elif api_struct.connection_info == MosaikSyncInfo:
        con_info = get_mosaik_connnection_info()
        collection_info = create_collection_info(api_struct.api, con_info)
        sync_info = MosaikSyncInfo(con_info, collection_info)
    else:
        raise NotImplementedError(f"TODO {api_struct}")
    api = api_struct.api(sync_info)
    if has_dup_idents(api, sync_info.collection_info):
        raise KeyError("Not a valid identifier")
    cfg_access.write_basic_connection(
        args.target, connection_type=args.source, sync_info=sync_info
    )
    logging.info("Verbindungsdaten wurden gespeichert.")
    return collection_info


def retrive_cfg_and_wattro_api() -> tuple[SyncCfg, WattroNodeApi]:
    cfg_existed = wattro_sync.file_access.read_write.exists("cfg")
    cfg = wattro_sync.config_reader.access.get_or_create()
    wattro_api = get_wattro_api_or_none(cfg)
    if wattro_api is None:
        raise ConnectionError("Failed to connect to wattor api.")
    if not cfg_existed:
        if select(
            ["nein", "ja"], required=True, title="E-Mail Benachrichtigung einrichten?"
        ):
            setup_mail_cfg(cfg)
    return cfg, wattro_api


def select_src(key: str, src_fields: list[str], value: dict) -> str | None:
    can_multi_select = value["type"] == "string"
    src = (
        _select_one(key, src_fields, value)
        if not can_multi_select
        else _select_for_string(key, src_fields, value)
    )
    logging.info("%s <-- %s", key, src)
    return src


def _select_for_string(key: str, src_fields: list[str], value: dict) -> str | None:
    required = value["required"]
    choices = multi_select(
        src_fields,
        required=required,
        title=f"Felder für {key} [Typ: string]{_skip_text(required)}",
    )
    if not choices:
        return None
    cfg_fields = [_make_cfg_field(src_fields[i]) for i in choices]
    return ", ".join(cfg_fields)


def _select_one(key: str, src_fields: list[str], value: dict) -> str | None:
    required = value["required"]
    choice = select(
        src_fields,
        required=required,
        title=f"Feld für {key} [Typ: {value['type']}]{_skip_text(required)}",
    )
    if choice is None:
        return None
    return _make_cfg_field(src_fields[choice])


def _skip_text(required: bool) -> str:
    if required:
        return "\t(Pflichtfeld)"
    return "\t(Überspringen: q)"


def _make_cfg_field(field_name: str) -> str:
    return "{" + field_name + "}"


def get_wattro_api_or_none(cfg) -> WattroNodeApi | None:
    logging.info("Prüfe Wattro API Verfügbarkeit.")
    try:
        wattro_api = WattroNodeApi.get_healthy_api(
            domain=cfg.wattro_cfg.domain, api_key=cfg.wattro_cfg.api_key
        )
    except ConnectionError as err:
        logging.critical(
            "Verbindung zu Wattro API fehlgeschlagen: %s",
            err,
        )
        return None
    logging.info("Wattro API erreichbar unter %s", wattro_api.hostname)
    return wattro_api


def get_mosaik_connnection_info() -> str:
    server_path = input("Server Pfad (z.B. SERVERNAME\\SQLMOSER):\t")
    db = input("Datenbank Name (z.B. Mosaik):\t")
    con_info = MosaikConInfo(Server=server_path, Database=db).to_connection_str()
    return con_info


def get_sqlite_connection_info() -> str:
    while True:
        db_path_str = input("Pfad zur Datenbank Datei (z.B. /path/to/db.sqlite3):\t")
        connection_info = pathlib.Path(db_path_str).resolve()
        if connection_info.is_file():
            return db_path_str
        logging.error(
            " '%s' muss ein Pfad zu einer gültigen Datei sein.", connection_info
        )


def has_dup_idents(api: SrcCli, collection_info: CollectionInfo) -> bool:
    res = api.get_new([])
    known_identifiers: dict[str, list[dict]] = dict()
    dups = list()
    for data in res:
        ident = data[collection_info.ident]
        if ident in known_identifiers:
            dups.append(ident)
            known_identifiers[ident].append(data)
        else:
            known_identifiers[ident] = [data]
    if not dups:
        return False

    for dup in dups:
        logging.error("%s ist nicht eindeutig: %s", collection_info.ident, dup)
        for dat in known_identifiers[dup]:
            logging.info("\t%s", dat)
    return True


def create_collection_info(
    api: typing.Type[SrcCli], connection_info: typing.Any
) -> CollectionInfo:
    table_options = api.get_collections(connection_info)
    table_id = select(
        table_options, required=True, title="Quelltabelle (Enter zum Auswählen)"
    )
    table_name = table_options[table_id]
    field_options, field_samples = api.get_fields(connection_info, table_name)
    field_ids = multi_select(
        field_options,
        required=True,
        title="Felder, die verwendet werden sollen",
        preview_title="Beispiel Werte",
        preview_command=lambda field_name: ", ".join(
            [str(x) for x in field_samples[field_name]]
        ),
    )
    field_choices = [
        field for idx, field in enumerate(field_options) if idx in field_ids
    ]
    identifier = field_choices[
        select(
            field_choices,
            required=True,
            title="Primärschlüssel (Feld, welches den Datensatz eindeutig identifiziert)",
        )
    ]
    collection_info = CollectionInfo(table_name, field_choices, identifier)
    return collection_info


def setup_mail_cfg(cfg: SyncCfg) -> None:
    """setup mail cfg, store it to file and cfg obj"""
    log_level_choices = {
        "Bei jedem Aufruf.": logging.DEBUG,
        "Wenn versucht wurde Daten zu schreiben.": logging.INFO,
        "Wenn Daten nicht geschrieben werden konnten.": logging.WARNING,
    }

    choice_idx = select(
        list(log_level_choices.keys()),
        required=True,
        title="Wann soll eine Mail gesendet werden?",
    )
    log_level = list(log_level_choices.values())[choice_idx]

    mail_cfg = MailCfg(
        api_key=input("Sendgrid Api-Key:\t"),
        to_emails=input("Empfänger E-Mail:\t"),
        log_level=log_level,
    )
    mail_api = MailApi(mail_cfg)
    if mail_api.send_registered():
        cfg.mail_cfg = mail_cfg
        cfg_access.write_mail_cfg(mail_cfg)
        logging.info(
            "Aktivierungs Mail erfolgreich verschickt und Konfiguration gespeichert."
        )
    else:
        logging.error("Mail Konfiguration verworfen.")


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()

    parser.add_argument("target", help="Daten Ziel", choices=TARGET_NODE_MAPPING.keys())
    parser.add_argument("source", help="Quellsystem", choices=SOURCE_CHOICES)

    return parser.parse_args()


if __name__ == "__main__":
    exit(main())



import hashlib
import json
from typing import Literal, NewType, Iterable, Iterator

from ..file_access import read_write

Target = Literal["asset", "project"]

HashHistory = NewType("HashHistory", dict[str, str])


class HistoryHandler:
    def __init__(self):
        self.full_hist = read_write.read("history")

    def iter_changed(
        self, target: str, to_check: Iterable[dict], ident: str
    ) -> Iterator[dict]:
        ident = ident
        hist = self.full_hist.get(target, {})
        for val in to_check:
            key = str(val[ident])
            if hist.get(key, None) != _hashed(val):
                yield val

    def update(self, target: str, val: dict, ident: str) -> None:
        ident = ident
        key = str(val[ident])
        if target not in self.full_hist:
            self.full_hist[target] = {}
        self.full_hist[target][key] = _hashed(val)

    def save(self) -> None:
        read_write.write("history", self.full_hist)


def _generate_from_values(new_values: Iterable[dict], ident: str) -> HashHistory:
    """_generate_from_values a hash history from new values, using ident as the key"""
    return HashHistory({str(val[ident]): _hashed(val) for val in new_values})


def _hashed(raw_dict: dict) -> str:
    """generate hash value from a dict that can be json serialized."""
    # sort_keys: make stable under dict shuffle
    seed = json.dumps(
        {key: f"{val}" for key, val in raw_dict.items()}, sort_keys=True
    ).encode()
    hashed = hashlib.md5(seed, usedforsecurity=False)
    return hashed.hexdigest()



#!/bin/env python3
import argparse
import dataclasses
import datetime
import logging
import random
from typing import Iterable, Sequence

import wattro_sync.config_reader.access
import wattro_sync.file_access.read_write
from wattro_sync.api.api_mapping import ApiNameToStructureMapping
from wattro_sync.api.mail import MailApi
from wattro_sync.api.rest_api import WattroNodeApi
from wattro_sync.api.src_cli import SrcCli
from wattro_sync.config_reader.types import (
    SyncCfg,
    ConnectionStructure,
    FieldMapping,
    ConfigDegenerated,
)
from wattro_sync.file_access.logging import get_file_handler, get_stdout_handler
from wattro_sync.hash_history import history
from wattro_sync.helpers import SOURCE_CHOICES, TARGET_NODE_MAPPING


def main() -> int:
    args = parse_args()

    setup_logger(args)
    if not wattro_sync.file_access.read_write.exists("cfg"):
        logging.error(
            "Keine Konfiguration gefunden. Bitte zuerst eine via `setup` erstellen."
        )
        return -1
    cfg: SyncCfg = wattro_sync.config_reader.access.get_or_create()
    mail_api = MailApi(cfg.mail_cfg)

    logging.info("Prüfe Verbindung zu Wattro.")
    try:
        wattro_api = WattroNodeApi.get_healthy_api(**dataclasses.asdict(cfg.wattro_cfg))
    except Exception as err:
        mail_api.send(
            logging.CRITICAL,
            "Verbindung zu Wattro API fehlgeschlagen. Es wurden keine Daten gesendet.",
        )
        logging.critical(err)
        return -1
    logging.info("Verbindung zu Wattro ok.")

    tot_success, tot_fail = 0, 0
    for target in TARGET_NODE_MAPPING:
        if args.limit_target and target not in args.limit_target:
            logging.info(f"Überspringe {target!r}, da nicht in {args.limit_target}.")
            continue
        connection_struct: ConnectionStructure | None = getattr(cfg, target, None)
        if connection_struct is None:
            logging.info(f"Überspringe {target!r}, da keine Konfiguration vorhanden.")
            continue
        con_type = connection_struct.connection_type
        if args.limit_src and con_type not in args.limit_src:
            logging.info(
                f"Überspringe {target}, da {con_type} nicht in {args.limit_src}."
            )
            continue

        success, fail = sync(target, connection_struct, wattro_api, args.dry)
        tot_success += success
        tot_fail += fail

    mail_log_lvl = logging.DEBUG
    tot = tot_fail + tot_success
    msg = "Synchronisation durchgeführt.<br /><br />"
    if tot > 1:
        msg += f"Es wurden {tot} Datensätze bearbeitet.<br />"
    else:
        msg = "Es wurde ein Datensatz bearbeitet.<br />"
    if tot_success > 0:
        mail_log_lvl = logging.INFO
        msg += f"Davon erfolgreich: {tot_success} ({tot_success / tot:.1%}).<br />"
    if tot_fail > 0:
        mail_log_lvl = logging.ERROR
        if tot_fail > 1:
            msg += f"Es traten <strong>{tot_fail} Fehler</strong> auf.<br />"
        else:
            msg += f"Dabei trat <strong>ein Fehler</strong> auf.<br />"

    if args.dry:
        msg += "Da es sich um einen Testaufruf handelt wurden <emph>keine Daten verändert</emph>."

    mail_api.send(mail_log_lvl, msg)

    logging.info("Sync beendet.")
    return 0


def setup_logger(args):
    base_logger = logging.getLogger()
    base_logger.setLevel(logging.INFO)
    file_handler, stdout_handler = get_file_handler(), get_stdout_handler()
    base_logger.addHandler(file_handler)
    base_logger.addHandler(stdout_handler)

    stdout_handler.setLevel(logging.WARNING)
    if args.v or args.dry:
        stdout_handler.setLevel(logging.INFO)
    logging.info(" === Started %s  ===", datetime.datetime.now())
    if args.dry:
        logging.info("Dry Run.")


def sync(
    target: str,
    src_con_struct: ConnectionStructure,
    wattro_api: WattroNodeApi,
    is_dry_run: bool,
) -> tuple[int, int]:
    """returns number of successfull and failed updates"""
    logging.info(
        f"Starte Prozess für {target!r} (Quelle: {src_con_struct.connection_type})"
    )
    source_api_struct = ApiNameToStructureMapping[src_con_struct.connection_type]
    api_class: type[SrcCli] = source_api_struct.api
    try:
        src_api: SrcCli = api_class.get_healthy_connection(src_con_struct.sync_info)
    except ConnectionError:
        logging.error("Prozess für %s abgebrochen.", target)
        return 0, 1
    logging.info("Hole Daten von Wattro...")
    known_idents = wattro_api.get_idents(target)

    logging.info("%s gefunden. Hole neue Daten von Quelle...", len(known_idents))
    hist = history.HistoryHandler()
    ident = src_con_struct.sync_info.collection_info.ident
    src_data_new_idents = src_api.get_new(known_idents)
    logging.info("%s neue gefunden.", len(src_data_new_idents.rows))

    success_updates = 0
    failed_updates = 0
    success = send_to_wattro(
        target, src_con_struct, wattro_api, is_dry_run, src_data_new_idents
    )
    if success and not is_dry_run:
        for val in src_data_new_idents:
            hist.update(target, val, ident)
        hist.save()
    if success:
        success_updates += len(src_data_new_idents)
    else:
        failed_updates += len(src_data_new_idents)

    logging.info("Hole Daten von Quelle...")
    src_data_known_idents = src_api.get_old(known_idents)
    logging.info("%s Datensätze auf Änderung prüfen...", len(src_data_known_idents))

    for changed in hist.iter_changed(target, src_data_known_idents, ident):
        success = send_to_wattro(
            target,
            src_con_struct,
            wattro_api,
            is_dry_run,
            src_data=[changed],
            update=True,
        )
        if success:
            success_updates += 1
            if not is_dry_run:
                hist.update(target, changed, ident)
        else:
            failed_updates += 1
    logging.info(
        f"Sync für %s abgeschlossen. Bearbeitet: %i (erfolgreich: %i | nicht erfolgreich: %i)",
        target,
        success_updates + failed_updates,
        success_updates,
        failed_updates,
    )
    if not is_dry_run and success_updates > 0:
        hist.save()
    return success_updates, failed_updates


def send_to_wattro(
    target: str,
    src_con_struct: ConnectionStructure,
    wattro_api: WattroNodeApi,
    is_dry_run: bool,
    src_data: Sequence[dict],
    update=False,
) -> bool:
    """send to wattro. return Fail if write failed"""
    if len(src_data) == 0:
        return True
    new_target_data = transform(
        src_data, src_con_struct.field_mapping, src_con_struct.encoding
    )
    if is_dry_run:
        count = len(new_target_data)
        logging.info("DRY RUN - Es wurden %i Datensätze erzeugt.", count)
        logging.info("Beispiel Datensätze (quelle --> ziel):")
        k = min(count, 3)
        for i in random.sample(range(count), k=k):
            logging.info("%s --> %s", src_data[i], new_target_data[i])
        return True
    logging.info("Schreibe Daten nach Wattro.")
    success = True
    try:
        if not update:
            wattro_api.bulk_create(target, new_target_data)
        else:
            wattro_api.update_by_ident(target, new_target_data[0])
    except ConnectionError as issue:
        logging.error("Schreiben von %s fehlgeschlagen. %s", new_target_data, issue)
        success = False
    logging.info("abgeschlossen.")
    return success


def transform(
    new_src_data: Iterable[dict], field_mapping: FieldMapping, encoding: str
) -> list[dict]:
    new_data_list = []
    for raw_src in new_src_data:
        if not isinstance(raw_src, dict):
            raise RuntimeError(f"Wrong instance format: {raw_src!r}")
        src = parse_raw_src(raw_src, encoding)
        new_data = {}
        for field_name, field_map in field_mapping.items():
            src_str = field_map["src"]
            if src_str is None:
                continue
            if isinstance(src_str, int):
                raise RuntimeError(
                    "Fehlkonfiguration Feld: {field_name}. Quelle ist fester Wert."
                )
            new_data[field_name] = get_date(field_map, field_name, src, src_str)
        new_data_list.append(new_data)
    return new_data_list


def parse_raw_src(raw_src: dict, encoding: str) -> dict[str, int | str]:
    src: dict[str, int | str] = {}
    for key, val in raw_src.items():
        if val is None:
            src[key] = ""
        elif isinstance(val, bytes):
            try:
                str_val = val.decode(encoding=encoding)
            except UnicodeDecodeError as decode_err:
                logging.warning(decode_err.reason, key, val)
                str_val = val.decode(encoding=encoding, errors="ignore")
            src[key] = str_val
        else:
            src[key] = val
    return src


def get_date(field_map: dict, field_name: str, src: dict, src_str: str) -> int | str:
    try:
        new_date = src_str.format(**src)
    except Exception as eval_fail:
        logging.critical(f"Failed to parse %r on %s for %s", src_str, src, field_name)
        raise eval_fail
    if field_map["type"] in ["field", "int"]:
        return int(new_date)

    max_length = field_map.get("max_length", None)
    if max_length is None:
        return new_date

    if not isinstance(max_length, int):
        raise ConfigDegenerated(f"max_length not a number {max_length=} {field_name=}")

    oneline = new_date.strip().replace("\r", "").replace("\n", " | ")

    if len(oneline) <= max_length:
        return oneline

    shortened = f"{oneline[:max_length - 3]}..."
    logging.info(
        "Eingabe für %r zu groß. Kürze %r ----> %r", field_name, new_date, shortened
    )
    return shortened


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--dry",
        action="store_true",
        help="Dry Run. Modifiziert keine Daten. Impliziert -v",
        default=False,
    )
    parser.add_argument(
        "--limit_src",
        help="Schränkt die Quellsysteme ein.",
        choices=SOURCE_CHOICES,
        nargs="*",
    )
    parser.add_argument(
        "--limit_target",
        help="Schränkt die Ziele ein.",
        choices=TARGET_NODE_MAPPING.keys(),
        nargs="*",
    )
    parser.add_argument("-v", help="Setzt das Loglevel auf 'info'", action="store_true")
    return parser.parse_args()


if __name__ == "__main__":
    exit(main())



import copy
import random
import unittest
import unittest.mock

from wattro_sync.hash_history import history
from wattro_sync.hash_history.history import HashHistory


class FileAccessMock(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_read_write = unittest.mock.patch(
            "wattro_sync.hash_history.history.read_write"
        ).start()
        self.mock_read_write.get_or_create.return_value = ("mock_path", False)
        self.mock_read_write.read.return_value = {}

    def tearDown(self) -> None:
        unittest.mock.patch.stopall()

    @staticmethod
    def gen_mock_hist(prefix: str = "a", entries: int = -1) -> HashHistory:
        if entries < 0:
            entries = random.randint(1, 10)
        return HashHistory(
            {f"{prefix}{i}": i.to_bytes(1, byteorder="big") for i in range(entries)}
        )

    @staticmethod
    def gen_list_of_dicts(
        keys: list[str] | None = None, entries: int = -1
    ) -> list[dict]:
        if keys is None:
            keys = [f"k{i}" for i in range(random.randint(1, 3))]
        if entries < 0:
            entries = random.randint(1, 10)
        result = []
        for i in range(entries):
            mapping = {key: random.choice([f"value{i}", i]) for key in keys}
            result.append(mapping)
        return result

    @staticmethod
    def get_ident(list_of_dict: list[dict]) -> str:
        return random.choice(list(list_of_dict[0].keys()))

    def gen_list_and_ident(self) -> tuple[list[dict], str]:
        list_of_dicts = self.gen_list_of_dicts()
        ident = self.get_ident(list_of_dicts)
        assert isinstance(ident, str)
        return list_of_dicts, ident


class TestHash(FileAccessMock):
    def test_get_empty(self) -> None:
        self.assertEqual({}, history._generate_from_values([], "key"))

    def test_get_history(self) -> None:
        list_of_dicts, ident = self.gen_list_and_ident()

        hist = history._generate_from_values(list_of_dicts, ident)

        self.assertEqual(len(hist), len(list_of_dicts))
        ident_list = [str(val[ident]) for val in list_of_dicts]
        for ident in hist.keys():
            self.assertIn(ident, ident_list)

    def test_time_stable_history(self) -> None:
        list_of_dicts, ident = self.gen_list_and_ident()

        hist1 = history._generate_from_values(list_of_dicts, ident)
        self.assertEqual(len(hist1), len(list_of_dicts))  # instead of 'sleep'
        hist2 = history._generate_from_values(list_of_dicts, ident)

        self.assertEqual(hist1, hist2)

    def test_list_sort_stable(self) -> None:
        list_of_dicts, ident = self.gen_list_and_ident()

        hist1 = history._generate_from_values(list_of_dicts, ident)
        random.shuffle(list_of_dicts)
        hist2 = history._generate_from_values(list_of_dicts, ident)

        self.assertEqual(hist1, hist2)

    def test_dict_sort_stable(self) -> None:
        list_of_dicts, ident = self.gen_list_and_ident()

        hist1 = history._generate_from_values(list_of_dicts, ident)
        shuffled_dicts = []
        for dict_ in list_of_dicts:
            keys = list(dict_.keys())
            random.shuffle(keys)
            shuffled_dicts.append({k: dict_[k] for k in keys})
        hist2 = history._generate_from_values(shuffled_dicts, ident)

        self.assertEqual(hist1, hist2)


class TestHistoryHandler(FileAccessMock):
    def setUp(self) -> None:
        super().setUp()
        self.mock_read_write.read.return_value = {"asset": self.gen_mock_hist()}

    def test_reads_on_create(self) -> None:
        history.HistoryHandler()
        self.mock_read_write.read.assert_called_once()

    def test_flush(self) -> None:
        old_file = self.mock_read_write.read()
        hh = history.HistoryHandler()
        hh.save()
        self.mock_read_write.write.assert_called_once_with("history", old_file)

    def test_udpate(self) -> None:
        old_file = copy.deepcopy(self.mock_read_write.read())
        hh = history.HistoryHandler()
        val_list, ident = self.gen_list_and_ident()
        val = val_list[0]
        target = list(old_file.keys())[0]
        hh.update(target, val, ident)
        hh.save()
        self.mock_read_write.write.assert_called_once()
        _, write_value = self.mock_read_write.write.mock_calls[0].args
        self.assertNotEqual(old_file, write_value)
        ident_val = str(val[ident])
        self.assertEqual(write_value[target][ident_val], history._hashed(val))


class TestIterChanged(FileAccessMock):
    def gen_list_ident_and_selfhist(self) -> tuple[list[dict], str, HashHistory]:
        to_check, ident = self.gen_list_and_ident()
        hist = history._generate_from_values(to_check, ident)
        return to_check, ident, hist

    def get_handler_with_hist(
        self, hist_file: dict, target: str = "target"
    ) -> history.HistoryHandler:
        self.mock_read_write.read.return_value = {target: hist_file}
        return history.HistoryHandler()

    def assertOneChanged(
        self,
        ident: str,
        ident_val: str,
        to_check: list[dict],
        handler: history.HistoryHandler,
        target: str = "target",
    ) -> None:
        try:
            expected = next(filter(lambda val: str(val[ident]) == ident_val, to_check))
        except StopIteration:
            self.fail(
                f"Mock setup error. Value with {ident_val = } for {ident = } in \n"
                f"{to_check = }\n"
            )
        changed_generator = handler.iter_changed(target, to_check, ident)
        try:
            next_dict = next(changed_generator)
        except StopIteration:
            self.fail(
                f"Unexpectetly no changes found for {target = }\n"
                f"{ident = }\n"
                f"{to_check = }\n"
                f"{handler.full_hist = }\n"
            )
        else:
            self.assertEqual(expected, next_dict)
        with self.assertRaises(StopIteration):
            next(changed_generator)

    def test_get_changed_no_init_data(self) -> None:
        hh = self.get_handler_with_hist({})
        to_check, ident = self.gen_list_and_ident()
        self.assertEqual(to_check, list(hh.iter_changed("target", to_check, ident)))

    def test_get_nothing_on_same(self) -> None:
        to_check, ident, hist = self.gen_list_ident_and_selfhist()
        hh = self.get_handler_with_hist(hist)
        self.assertEqual([], list(hh.iter_changed("target", to_check, ident)))

    def test_get_changed_correct_target(self) -> None:
        to_check, ident, hist = self.gen_list_ident_and_selfhist()
        hh = self.get_handler_with_hist(hist)
        self.assertEqual(
            to_check, list(hh.iter_changed("other target", to_check, ident))
        )

    def test_finds_missing(self) -> None:
        to_check, ident, hist = self.gen_list_ident_and_selfhist()

        ident_val = random.choice(list(hist.keys()))
        hist.pop(ident_val)
        hh = self.get_handler_with_hist(hist)

        self.assertOneChanged(ident, ident_val, to_check, hh)

    def test_finds_changed(self) -> None:
        to_check, ident, hist = self.gen_list_ident_and_selfhist()

        ident_val = random.choice(list(hist.keys()))
        hist[ident_val] += "xxx"
        hh = self.get_handler_with_hist(hist)

        self.assertOneChanged(ident, ident_val, to_check, hh)


if __name__ == "__main__":
    unittest.main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="wattro_sync",
    version="0.3.1",
    description="Script collection to sync data from local sources to a wattro node",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wattro/wattro_sync",
    author="Wattro GmbH",
    author_email="admin@wattro.de",
    packages=setuptools.find_packages(),
    license="MIT License",
    install_requires=["requests", "sendgrid", "simple-term-menu", "pyodbc"],
)



from __future__ import annotations

import dataclasses
from abc import ABC

import pyodbc as pyodbc

from wattro_sync.api.src_cli import SyncInfo, CollectionInfo, SrcCli, DBRes


class OdbcSyncInfo(SyncInfo):
    odbc_connection_str: str
    collection_info: CollectionInfo

    def __init__(self, odbc_connection_str: str, collection_info: CollectionInfo):
        self.odbc_connection_str = odbc_connection_str
        self.collection_info = collection_info

    @classmethod
    def from_dict(cls, info: dict) -> OdbcSyncInfo:
        return cls(
            info["odbc_connection_str"], CollectionInfo(**info["collection_info"])
        )

    def asdict(self) -> dict:
        return {
            "odbc_connection_str": self.odbc_connection_str,
            "collection_info": dataclasses.asdict(self.collection_info),
        }


class OdbcSrcCli(SrcCli, ABC):
    def __init__(self, sync_info: OdbcSyncInfo):
        self.collection_info = sync_info.collection_info
        self.odbc_connection_str = sync_info.odbc_connection_str

    def _exec(self, qry: str, params=None) -> DBRes:
        try:
            cnxn = pyodbc.connect(self.odbc_connection_str)
        except Exception as err:
            raise ConnectionError(
                f"Failed to connect using {self.odbc_connection_str!r}"
            ) from err
        try:
            cursor = cnxn.cursor()
            if params is None:
                params = tuple()
            rows = cursor.execute(qry, params).fetchall()
            if not rows:
                res = DBRes([], [])
            else:
                res = DBRes([n[0] for n in cursor.description], rows)
        except Exception as err:
            raise RuntimeError(
                f"Failed to execute {qry = !r} with {params = !r}"
            ) from err
        finally:
            cnxn.close()
        return res


class OdbcSrcCliAltSample(OdbcSrcCli, ABC):
    def get_sample(self) -> DBRes:
        """For Sources that do not implement the LIMIT statement but the TOP statement."""
        qry = self._qry("").strip(";")
        sample_table = f"({qry}) as sample_table"
        return self._exec(f"SELECT TOP 1 * FROM {sample_table};")



from __future__ import annotations

import dataclasses
import logging
import pathlib

from .types import WattroCfg, ConnectionStructure, SyncCfg, MailCfg, ConfigDegenerated
from ..api.api_mapping import ApiNameToStructureMapping, ApiStructure
from ..api.src_cli import SyncInfo
from ..file_access.read_write import (
    CON_TYPE_KEY,
    CON_INFO_KEY,
    FIELD_MAP_KEY,
    cfg_update,
    write_path,
    get_or_create as get_or_create_fp,
    read_path,
)


def extract_wattro_cfg(cfg: dict) -> WattroCfg:
    """read_path wattro cfg form file"""
    if "wattro_cfg" not in cfg:
        raise ConfigDegenerated(f"'wattro_cfg' fehlt in {cfg}.")
    return WattroCfg(**cfg["wattro_cfg"])


def extract_src_cfg(cfg: dict, target: str) -> ConnectionStructure | None:
    """extract key from cfg and return structured data"""
    if target not in cfg:
        logging.info("Keine Konfiguration für %s. (Verfügbar: %s)", target, cfg.keys())
        return None
    cfg_data = cfg[target]
    con_type = cfg_data.get("connection_type", "")
    api_structure: ApiStructure = ApiNameToStructureMapping[con_type]

    con_info_data = cfg_data.get("connection_info", {})
    con_info = api_structure.connection_info.from_dict(con_info_data)
    field_mapping = cfg_data.get("field_mapping", {})
    encoding = cfg_data.get("encoding", "utf-8")

    return ConnectionStructure(
        connection_type=con_type,
        sync_info=con_info,
        field_mapping=field_mapping,
        encoding=encoding,
    )


def extract_mail_cfg(cfg: dict) -> MailCfg | None:
    """extract from dict and return structured data"""
    key = "mail"
    if key not in cfg:
        logging.info("Keine Konfiguration für E-Mail Benachrichtungen gefunden.")
        return None
    mail_cfg = cfg[key]
    return MailCfg(**mail_cfg)


def write_basic_connection(
    target: str, connection_type: str, sync_info: SyncInfo
) -> None:
    cfg_update(target, CON_TYPE_KEY, connection_type)
    cfg_update(target, CON_INFO_KEY, sync_info.asdict())


def write_field_mapping(target: str, field_mapping: dict) -> None:
    cfg_update(target, FIELD_MAP_KEY, field_mapping)


def _create_min_cfg(cfg_file: pathlib.Path) -> None:
    """Create Minimal Config."""
    print(f"Erzeuge Konfiguration in {cfg_file}. Bitte Informationen eingeben:")
    domain = input("Kunden Domain (Beispiel: `wattro`):\t")
    api_key = input("Api Key (Beispiel: AYpxtcVb.jYmvgB5A5DfQou6kSwS32LvmWUkUclo8):\t")
    wattro_cfg = WattroCfg(domain, api_key)
    minimal_config = {"wattro_cfg": dataclasses.asdict(wattro_cfg)}
    write_path(cfg_file, minimal_config)


def get_or_create() -> SyncCfg:
    """_get or create cfg"""
    cfg_file, created = get_or_create_fp("cfg")
    if created:
        _create_min_cfg(cfg_file)
    logging.info("Lese Konfiguration von %s", cfg_file)
    cfg = read_path(cfg_file)
    wattro_cfg = extract_wattro_cfg(cfg)
    asset_cfg = extract_src_cfg(cfg, "asset")
    project_cfg = extract_src_cfg(cfg, "project")
    mail_cfg = extract_mail_cfg(cfg)
    return SyncCfg(wattro_cfg, asset_cfg, project_cfg, mail_cfg)


def write_mail_cfg(mail_cfg: MailCfg) -> None:
    for key, value in dataclasses.asdict(mail_cfg).items():
        cfg_update("mail", key, value)



from __future__ import annotations

import abc
import logging
from dataclasses import dataclass
from typing import Any, Iterator, Sequence, overload


class SyncInfo(abc.ABC):
    collection_info: CollectionInfo

    @abc.abstractmethod
    def asdict(self) -> dict:
        ...

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, info: dict):
        ...


@dataclass
class CollectionInfo:
    collection_name: str
    fields: list[str]
    ident: str
    hardcoded_select: None | str = None

    @classmethod
    def empty(cls) -> CollectionInfo:
        return cls(collection_name="empty collection", fields=[], ident="empty")


@dataclass
class DBRes(Sequence):
    description: Sequence[str]
    rows: Sequence[tuple]

    def _dictify_row(self, row: tuple) -> dict:
        return {key: val for key, val in zip(self.description, row)}

    def __len__(self) -> int:
        return len(self.rows)

    def __iter__(self) -> Iterator[dict]:
        return (self._dictify_row(row) for row in self.rows)

    @overload
    def __getitem__(self, index: int) -> dict:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[dict]:
        ...

    def __getitem__(self, index: int | slice) -> dict | Sequence[dict]:
        if isinstance(index, int):
            return self._dictify_row(self.rows[index])
        return [self._dictify_row(row) for row in self.rows[index]]


class SrcCli(abc.ABC):
    collection_info: CollectionInfo

    def _qry(self, restrict: str) -> str:
        select = (
            self.collection_info.hardcoded_select
            or f"SELECT {','.join(self.collection_info.fields)} FROM {self.collection_info.collection_name}"
        )
        return f"{select} {restrict.strip(';')};"

    @classmethod
    def get_healthy_connection(cls, sync_info: SyncInfo):
        inst = cls(sync_info)
        try:
            sample = inst.get_sample()
        except Exception as exc_info:
            logging.error(f"Failed to get API: {exc_info}")
            raise ConnectionError("Failed to get API") from exc_info
        logging.debug(
            "Successfully sampled %s: %s",
            sync_info.collection_info.collection_name,
            sample,
        )
        return inst

    def get_sample(self) -> DBRes:
        """
        Get Sample entry
        """
        return self._exec(self._qry("LIMIT 1"))

    def get_new(self, known_idents: Sequence[str]) -> DBRes:
        """
        Get all entries on the CollectionInfo collection that are not identified by `known_idents`
        """
        restrict = ""
        if len(known_idents) > 0:
            restrict = f"WHERE {self.collection_info.ident} NOT IN ({','.join(['?' for _ in known_idents])})"
        qry = self._qry(restrict)
        return self._exec(qry, known_idents)

    def get_old(self, known_idents: Sequence[str]) -> DBRes:
        if len(known_idents) == 0:
            return DBRes([], [])
        restrict = f"WHERE {self.collection_info.ident} IN ({','.join(['?' for _ in known_idents])})"
        qry = self._qry(restrict)
        return self._exec(qry, known_idents)

    @abc.abstractmethod
    def __init__(self, sync_info: SyncInfo):
        ...

    @abc.abstractmethod
    def _exec(self, qry: str, params=None) -> DBRes:
        ...

    @classmethod
    @abc.abstractmethod
    def get_collections(cls, connection_info: Any) -> list[str]:
        """
        Get a list of collections in the src
        """
        ...

    @classmethod
    @abc.abstractmethod
    def get_fields(
        cls, connection_info: Any, collection: str
    ) -> tuple[list[str], dict[str, list]]:
        """
        get_fields(%connection_info%, 'CollectionToScan')
        returns a tuple of
        - list of field names (str)
        - list of a list of up to 10 sample values
        """
        ...
